package transaction;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;

import transaction.LogEntry.LogType;


/**
 * A simple log manager which logs operations on a ResourceManager
 * in memory.
 * @author bross
 *
 */
public class LogManager {
	
	private Hashtable<Integer, LinkedList<LogEntry>> log;
	private File workingDir;
	
	
	public LogManager(String workingDir) throws IOException{
		log = new Hashtable<Integer, LinkedList<LogEntry>>();
		this.workingDir = new File(workingDir);
		if(!this.workingDir.isDirectory())
			throw new IOException("the given working dir is not a directory: " + workingDir);
		
		readPersistedLogs();
	}
	
	public boolean log(int xid, String resourceID, LogType type,
			ResourceData afterImage, String reservation){
		
		LogEntry u = new LogEntry(xid, resourceID, type, afterImage, reservation);
		LinkedList<LogEntry> logEntries = log.get(xid);
		if(logEntries == null){
			logEntries = new LinkedList<LogEntry>();
			log.put(xid, logEntries);
		}
				
		logEntries.add(u);
		
		return true;
	}		
	
	public LinkedList<LogEntry> getLogEntries(int xid){
		return log.get(xid);
	}
	
	public void persistLogs(int xid) throws IOException{
		File f = new File(workingDir, xid + ".log");
		ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(f));
		oos.writeInt(xid);
		oos.writeObject(log.get(xid));
		oos.flush();
		oos.close();
	}
	
	@SuppressWarnings("unchecked")
	private void readPersistedLogs() throws IOException{
		File[] logFiles = workingDir.listFiles(new FilenameFilter() {
			
			@Override
			public boolean accept(File dir, String name) {
				return name.endsWith("log");
			}
		});
		try {
			for (File file : logFiles) {
				ObjectInputStream ois = new ObjectInputStream(new FileInputStream(file));
				int xid = ois.readInt();
				LinkedList<LogEntry> logitems = (LinkedList<LogEntry>)ois.readObject();
				log.put(xid, logitems);
				ois.close();
			}
		} catch (ClassNotFoundException e) {
			throw new IOException(e);
		}
	}
	
	public void deleteLogsForTransaction(int xid){
		log.remove(xid);
		File f = new File(workingDir, String.valueOf(xid));
		if(f.exists())
			f.delete();
	}	
	
	public ResourceData insertedResource(int xid, String rid){
		LinkedList<LogEntry> logEntries = log.get(xid);
		if(logEntries != null){
			for (LogEntry logEntry : logEntries) {
				if(logEntry.resourceID.equals(rid) && logEntry.type == LogType.INSERT_RESOURCE)
					return logEntry.afterImage;
			}
		}
		
		return null;
	}		
	
	public boolean insertedCustomer(int xid, String rid){
		LinkedList<LogEntry> logEntries = log.get(xid);
		if(logEntries != null){
			for (LogEntry logEntry : logEntries) {
				if(logEntry.resourceID.equals(rid) && logEntry.type == LogType.INSERT_CUSTOMER)
					return true;
			}
		}
		
		return false;
	}		
	
	public Set<String> getReservations(int xid, String customer){
		Set<String> reservations = new HashSet<String>();
		LinkedList<LogEntry> logEntries = log.get(xid);
		if(logEntries != null){
			for (LogEntry logEntry : logEntries) {
				if(logEntry.resourceID.equals(customer) && logEntry.type == LogType.INSERT_RESERVATION)
					reservations.add(logEntry.reservation);
			}
		}
		
		return reservations;		
	}
	
	public ResourceData getLatestUpdate(int xid, String rid){
		LinkedList<LogEntry> logEntries = log.get(xid);
		LogEntry entry = null;
		if(logEntries != null){
			Iterator<LogEntry> it = logEntries.descendingIterator();
			while(it.hasNext()){
				entry = it.next();
				if(entry.type == LogType.UPDATE_RESOURCE && entry.xid == xid && entry.resourceID.equals(rid))
					return entry.afterImage;
			}
		}
		
		return null;
	}
	
	public Map<String, ResourceData> getLatestResourceUpdates(int xid){
		HashMap<String, ResourceData> updates = new HashMap<String, ResourceData>();
		LinkedList<LogEntry> logEntries = log.get(xid);
		if(logEntries != null){
			for (LogEntry logEntry : logEntries) {
				if(logEntry.type == LogType.INSERT_RESOURCE || logEntry.type == LogType.UPDATE_RESOURCE)
					updates.put(logEntry.resourceID, logEntry.afterImage);
			}
		}
		
		return updates;
		
	}
	

}
